from __future__ import annotations

import copy
import yaml
import glob
from shutil import which
from typing import Any, Union
from pathlib import Path

import tricot.utils
from tricot.docker import TricotContainer
from tricot.logging import Logger
from tricot.validation import Validator, ValidationException
from tricot.extractor import Extractor
from tricot.plugin import Plugin
from tricot.command import Command
from tricot.condition import Condition


skip_until = None
assigned_ids = set()


class DuplicateIDError(Exception):
    '''
    Custom exception that is raised if two testers/tests use the same ID.
    '''


class TricotException(Exception):
    '''
    Custom exception class for general purpose and tricot related exceptions.
    '''
    def __init__(self, message: str, path: Path = None) -> None:
        '''
        Basically the default Exception class initializer, but accepts also the path
        to a configuration file.
        '''
        self.path = str(path.resolve())
        super().__init__(message)


class TricotRequiredFile(Exception):
    '''
    Custom exception class for a missing required file.
    '''


class TricotRequiredCommand(Exception):
    '''
    Custom exception class for a missing required command.
    '''


class TricotVersionMismatch(Exception):
    '''
    Custom exception class when a mismatching tricot version was used.
    '''


class ExceptionWrapper(Exception):
    '''
    Custom exception class that wraps around the actual exception and
    allows storing additional information.
    '''
    def __init__(self, original: Exception, path: Path = None) -> None:
        '''
        Store the original exception as well as the path of the file that
        caused it.
        '''
        self.path = str(path.resolve())
        self.original = original
        super().__init__(str(original))


class TesterKeyError(Exception):
    '''
    TesterKeyErrors are raised when the .yml test definition file misses
    required keys for constructing a Tester object.
    '''

    def __init__(self, key: str, path: Path, optional: str = None, section: str = None) -> None:
        '''
        As these Exception is only used internally, we give it a static content
        that only takes the missing key as argument.
        '''
        if optional is not None:
            self.message = f"Test configuration requires either the '{key}' or the '{optional}' key"
            self.message += "within the 'global' section."

        elif section is not None:
            self.message = f"Test configuration misses required key {key} in the '{section}' section."

        else:
            self.message = f"Test configuration misses required key {key} in the 'global' section."

        self.path = str(path.resolve())
        super().__init__(self.message)


class TestKeyError(Exception):
    '''
    TestKeyErrors are raised when the .yml test definition file misses
    required keys for constructing a Test object.
    '''

    def __init__(self, key: str, path: Path, message: str = None) -> None:
        '''
        As these Exception is only used internally, we give it a static content
        that only takes the missing key as argument. Optionally, we allow to add
        a custom message if required.
        '''
        if message:
            self.message = message

        else:
            self.message = f"Test configuration misses required key {key} within the "

        self.path = str(path.resolve())
        self.number = None

        super().__init__(self.message)

    def add_number(self, ctr: int) -> None:
        '''
        The default content of the exception indicates a missing key within a test configuration.
        This function can be used to add the position of the corresponding test within the .yml
        file.
        '''
        self.number = ctr

    def __str__(self) -> str:
        '''
        Depending whether the execption contains the test number (self.number), the exception
        is formatted a little bit differently.
        '''
        message = self.message

        if self.number:
            ordinal = tricot.utils.make_ordinal(self.number)
            message += f'{ordinal} test in the '

        return message + 'tests section.'


class Test:
    '''
    The Test class represents a single command test specified within a .yml file.
    Test objects contain all required information like the command definition, the
    specified variables and the required validators. During a test, Test objects are
    evaluated by executing their 'run' method.
    '''
    expected_keys = ['title', 'description', 'command', 'arguments', 'validators', 'timeout', 'env', 'conditions',
                     'logfile', 'shell', 'extractors', 'id', 'groups']

    def __init__(self, path: Path, title: str, error_mode: str, variables: dict[str, Any], command: Command,
                 timeout: int, validators: list[Validator], extractors: list[Extractor], env: dict, conditions: dict,
                 conditionals: set[Condition], test_id: str, test_groups: list[list[str]]) -> None:
        '''
        Initializer for a Test object.

        Parameters:
            path            Path object to the file that contains the tests configuration
            title           The title of the test - is displayed during execution.
            error_mode      Decides what to do if a validator fails (break|continue)
            variables       Dictionary of variables that are replaced in the command
            command         The command that is run by this test
            timeout         Timeout in seconds to wait for the specified command
            validators      List of validators to apply on the command output
            extractors      List of extractors to apply on the command output
            env             Environment variables
            conditions      Conditions for running the current test
            conditionals    Conditionals defined by upper testers
            test_id         Identifikation number of the test
            test_groups     Test groups that the test belongs to

        Returns:
            None
        '''
        self.path = path
        self.title = title
        self.error_mode = error_mode
        self.variables = variables
        self.command = command
        self.timeout = timeout
        self.validators = validators
        self.extractors = extractors
        self.env = env
        self.conditions = conditions
        self.conditionals = conditionals

        if test_id is None:
            self.id = self.title

        else:
            self.id = str(test_id)

            if self.id in assigned_ids:
                raise DuplicateIDError(f"ID '{self.id}' was assigned twice.")

            assigned_ids.add(self.id)

        self.groups = test_groups

        self.logfile = None
        self.success_string = 'success'
        self.failure_string = 'failed'
        self.success_color = 'green'
        self.failure_color = 'red'

    def set_logfile(self, logfile: str) -> None:
        '''
        Sets the logfile attribute on the Test object.

        Parameters:
            logfile         File system path to the logfile

        Returns:
            None
        '''
        if logfile is None:
            return

        logfile = tricot.utils.apply_variables(logfile, self.variables)
        self.logfile = open(logfile, 'w')

    def set_output(self, output_conf: dict) -> None:
        '''
        Parses an optional output dictionary that can be specified on test or tester level.
        The output dictionary can be used to overwrite default output values.

        Parameters:
            output_conf         Dictionary containing values like 'success_string' or 'success_color'

        Returns:
            None
        '''
        if not output_conf:
            return

        if type(output_conf) is not dict:
            raise TestKeyError('', 'When specified, output needs to be a dictionary.', self.path)

        self.success_string = output_conf.get('success_string') or self.success_string
        self.success_color = output_conf.get('success_color') or self.success_color

        self.failure_string = output_conf.get('failure_string') or self.failure_string
        self.failure_color = output_conf.get('failure_color') or self.failure_color

        tricot.utils.validate_color(self.success_color)
        tricot.utils.validate_color(self.failure_color)

    def apply_variables(val: Union(str, list), variables: dict[str, Any], k: str = 'command') -> list:
        '''
        Applies variables to a command or argument list/string. This method needs to be
        static, as it is applied right at the object creation. This is the only way to
        allow command lists to be specified within variables while simoultaneously checking
        types on object creation.

        Parameters:
            val             Value to replace variables in. Usually the command or arguments portion of .yml files
            variables       Variables to apply
            k               Current key value (command|arguments)

        Returns:
            None
        '''
        if val is None:
            return []

        for key, value in variables.items():

            value = tricot.utils.resolve_runtime_variables(variables, key, value)
            value = tricot.utils.resolve_env_variables(variables, key, value)
            key = '${'+str(key)+'}'

            if type(val) is str:

                if val == key and type(value) is list:
                    val = value.copy()

                else:
                    val = val.replace(key, str(value))

            elif type(val) is list:

                for ctr in range(len(val)):

                    if type(val[ctr]) is str:

                        if val[ctr] == key and type(value) is list:
                            val[ctr:ctr+1] = value

                        else:
                            val[ctr] = val[ctr].replace(key, str(value))

                    else:
                        val[ctr] = str(val[ctr])

        if type(val) is not list:
            raise ValueError(f"The '{k}' key needs to be a list within the ")

        return val

    def from_dict(path: Path, input_dict: dict, variables: dict[str, Any] = {}, error_mode: str = 'continue',
                  environment: dict = {}, conditionals: set[Condition] = set(), output_conf: dict = {},
                  parent_groups: list[list[str]] = list(), suggested_id: str = None) -> Test:
        '''
        Creates a Test object from a dictionary. The dictionary is expected to be the content
        read in of a .yml file and needs all keys that are required for a test (validators,
        title and command).

        Parameters:
            path            Path object to the tests configuration file
            input_dict      Dictionary that defines all required information for a test
            variables       Variables that should be inherited from the global scope
            error_mode      Decides whether to break or continue on failure
            environment     Environment variables
            conditionals    Conditionals for running the test
            output_conf     Output configuration inherited by the tester
            parent_groups   Test groups inherited from the parent tester
            suggested_id    Tets ID suggested by the id_pattern

        Returns:
            Test            Newly generated Test object
        '''
        try:
            j = input_dict
            var = tricot.utils.merge(variables, j.get('variables', {}), 'variables', path)
            validators = Validator.from_list(path, j['validators'], var)
            extractors = Extractor.from_list(path, j.get('extractors', []), var)
            groups = tricot.utils.merge_groups(parent_groups, list(map(lambda x: str(x), j.get('groups', []))))

            e_mode = j.get('error_mode') or error_mode
            env = tricot.utils.merge_environment(j.get('env'), environment, path)
            conditions = j.get('conditions', {})

            Condition.check_format(path, conditions, conditionals)

            command = Test.apply_variables(j['command'], var)
            arguments = Test.apply_variables(j.get('arguments'), var, 'arguments')

            if type(command) is not list:
                raise TestKeyError(None, path, "The 'command' key needs to be a list within the ")

            if arguments:

                if type(arguments) is not list:
                    raise TestKeyError(None, path, "The 'arguments' key needs to be a list within the ")

                command += arguments

            shell = j.get('shell', False)
            command = Command(command, shell)

            tricot.utils.check_keys(Test.expected_keys, input_dict)
            test = Test(path, j['title'], e_mode, var, command, j.get('timeout'), validators, extractors, env,
                        conditions, conditionals, j.get('id', suggested_id), groups)

            test.set_logfile(j.get('logfile'))
            test.set_output(tricot.utils.merge(output_conf, j.get('output', {}), 'output', path))

            return test

        except KeyError as e:
            raise TestKeyError(str(e), path)

        except ValueError as e:
            raise TestKeyError(None, path, str(e))

    def from_list(path: Path, input_list: list, variables: dict[str, Any] = {}, error_mode: str = 'continue',
                  env: dict = {}, conditionals: set[Condition] = set(), output_conf: dict = {},
                  parent_groups: list[list[str]] = list(), id_pattern: str = None, index: int = 0) -> list[Test]:
        '''
        Within .yml files, Tests are specified in form of a list. This function takes such a list,
        that contains each single test definition as another dictionary (like it is created when
        reading the .yml file) and returns a corresponding list of Test objects.

        Parameters:
            path            Path object to the tests configuration file
            input_list      List of test definitions as read in from a .yml file
            variables       Variables that should be inherited from the global scope
            error_mode      Decides whether to break or continue on failure
            env             Environment variables
            conditionals    Conditionals specified by the upper tester
            output_conf     Output configuration inherited by the tester
            parent_groups   Test groups inherited from the parent tester
            id_pattern      Pattern to create test IDs from
            index           Initial Index to use for the first test

        Returns
            list[Test]      List of Test objects created from the .yml input
        '''
        tests = list()

        if type(input_list) is not list:
            raise TestKeyError(None, path, 'Test defintions need to be specified as list within the ')

        for ctr in range(len(input_list)):

            if id_pattern is not None:
                suggested_id = id_pattern.format(ctr + index)

            else:
                suggested_id = None

            try:
                test = Test.from_dict(path, input_list[ctr], variables, error_mode, env, conditionals,
                                      output_conf, parent_groups, suggested_id)
                tests.append(test)

            except TestKeyError as e:
                e.add_number(ctr + 1)
                raise e

        return tests

    def from_file(path: Path, variables: dict[str, Any] = {}, error_mode: str = 'continue',
                  env: dict = {}, conditionals: set[Condition] = set(), output_conf: dict = {},
                  parent_groups: list[list[str]] = list(), id_pattern: str = None, index: int = 0) -> list[Test]:
        '''
        Reads a list of test definitions from a yaml file. Tests should be contained in a 'tests' section
        that contains each single test definition.

        Parameters:
            path            Path object to the tests configuration file
            variables       Variables that should be inherited from the global scope
            error_mode      Decides whether to break or continue on failure
            env             Environment variables
            conditionals    Conditionals specified by the upper tester
            output_conf     Output configuration inherited by the tester
            parent_groups   Test groups inherited from the parent tester
            id_pattern      Pattern to create test IDs from
            index           Initial Index to use for the first test

        Returns
            list[Test]      List of Test objects created from the .yml input
        '''
        with open(path, 'r') as config:

            try:
                config_dict = yaml.safe_load(config.read())

            except (yaml.parser.ParserError, yaml.scanner.ScannerError) as e:
                raise ExceptionWrapper(e, path)

        if 'tests' in config_dict and type(config_dict['tests']) is list:
            return Test.from_list(path, config_dict['tests'], variables, error_mode, env, conditionals, output_conf,
                                  parent_groups, id_pattern, index)

        return []

    def run(self, prefix: str = '-', hotplug_variables: dict[str, Any] = None) -> None:
        '''
        Runs the Test and applies all specified validators to the command output.
        Depending on the current error_mode, the function may raise exceptions and
        let the caller handle them. Furthermore, it generates output regarding the
        success or error of a Test.

        Parameters:
            prefix              Optional prefix to include for each test title
            hotplug_variables   Variables that are applied at runtime.

        Returns:
            None
        '''
        Logger.add_logfile(self.logfile)

        if self.id and self.id != self.title:
            Logger.print_blue(f'{prefix} [{self.id}] {self.title}...', end=' ', flush=True)
        else:
            Logger.print_blue(f'{prefix} {self.title}...', end=' ', flush=True)

        success = True

        if not Condition.check_conditions(self.conditions, self.conditionals):
            Logger.cprint('skipped.', color='grey')
            return

        self.command.run(self.path.parent, self.timeout, hotplug_variables, self.env)
        extractor_error = None

        for extractor in self.extractors:

            try:
                extractor._extract(self.command, hotplug_variables)

            except Exception as e:

                if extractor.on_miss == 'continue':
                    continue

                elif extractor.on_miss == 'warn':
                    Logger.extract_warning(e, extractor)
                    continue

                elif extractor.on_miss == 'break':
                    extractor_error = e
                    break

        for validator in self.validators:

            try:
                validator._run(self.command, hotplug_variables, extractor_error)

            except Exception as e:

                if success:
                    f_color = validator.failure_color or self.failure_color
                    f_string = validator.failure_string or self.failure_string
                    Logger.cprint(f_string, color=f_color)

                if extractor_error is not None:
                    tricot.constants.LAST_ERROR = tricot.constants.EXTRACT_EXCEPTION
                else:
                    tricot.constants.LAST_ERROR = tricot.constants.VALIDATION_EXCEPTION

                Logger.handle_error(e, validator)
                Condition.update_conditions(self.conditions, self.conditionals, True)

                if self.error_mode == "break":
                    Logger.remove_logfile(self.logfile)
                    Logger.decrease_indent()

                    if extractor_error:
                        raise extractor_error

                    raise ValidationException('')

                elif extractor_error is not None:
                    success = False
                    break

                else:
                    success = False

        if success:
            Condition.update_conditions(self.conditions, self.conditionals, False)
            Logger.cprint(self.success_string, color=self.success_color)
            Logger.handle_success(self.command, self.validators)

        hotplug_variables['$prev'] = self.command
        Logger.remove_logfile(self.logfile)

    def skip_test(self, exclude: set[str], exclude_groups: list[list[str]]) -> bool:
        '''
        Checks whether the current test is contained within the exclude lists.

        Parameters:
            exclude             Set of Test / Tester IDs to exclude
            exclude_groups      List of group sets to exclude

        Returns:
            bool
        '''
        if exclude and self.id in exclude:
            return True

        elif exclude_groups and tricot.utils.groups_contain(exclude_groups, self.groups):
            return True

        return False


class Tester:
    '''
    The Tester class represents a single .yml file that contains Test definitions. It is basically
    an Object that contains the tester configuration and a reference to each Test object that should
    be executed during a run. A command test with tricot is intended to be started by a Tester object,
    by calling its 'run' method.
    '''
    tester_count = 0

    def __init__(self, path: Path, title: str, variables: dict[str, Any], tests: list[Test], testers: list[Tester],
                 containers: list[TricotContainer], plugins: list[Plugin], conditions: dict, conditionals: set[Condition],
                 error_mode: str, tester_id: str, test_groups: list[list[str]], requires: dict) -> None:
        '''
        Initializes a new Tester object.

        Parameters:
            path            Path object to the Testers configuration file
            title           Title of the test (used for displaying)
            variables       Dictionary of global variables that should be applied to all Tests and Validators
            tests           List of Test objects that should run during a test
            testers         List of Tester object, that allows to nest tester objects
            containers      List of TricotContainer objects for the test
            plugins         List of Plugin objects for the test
            conditions      Conditions for running the current tester
            conditionals    Conditionals defined by the tester or upper testers
            error_mode      Decides what to do if a plugin fails (break|continue)
            tester_id       Unique identifikation number of the tester
            test_groups     Test groups that the tester belongs to
            requires        Requirements to run the tester

        Returns:
            None
        '''
        self.path = path
        self.title = title
        self.variables = variables
        self.tests = tests
        self.testers = testers
        self.containers = containers
        self.plugins = plugins
        self.conditions = conditions
        self.conditionals = conditionals
        self.error_mode = error_mode

        if tester_id is None:
            self.id = self.title

        else:
            self.id = str(tester_id)

            if self.id in assigned_ids:
                raise DuplicateIDError(f"ID '{self.id}' was assigned twice.")

            assigned_ids.add(self.id)

        self.groups = test_groups
        self.requires = requires

        self.logfile = None
        self.runall = False

    def from_dict(input_dict: dict, initial_vars: dict[str, Any] = dict(),
                  path: Path = None, e_mode: str = None, environment: dict = {},
                  conditionals: set[Condition] = set(), output_conf: dict = {},
                  test_groups: list[list[str]] = []) -> Tester:
        '''
        Creates a new Tester object from a python dictionary. The dictionary is expected to be
        created by reading a .yml file that contains test defintions. It requires all keys that
        need to be present for creating the Tester and the Test objects. This method walks down
        basically the complete .yml tree and constructs all required objects on its way.

        Parameters:
            input_dict      Python dictionary as read in from a .yml file
            initial_vars    Additional variables. This is used internally when nesting Testers
            path            Path object to the testers configuration file
            e_mode          Error mode that was mayve inherited by the parent tester
            environment     Dictionary of environment variables to use within the test
            conditionals    Conditions inherited from the upper tester
            output_conf     Output configuration inherited from the upper tester
            test_groups     List of test groups inherited from the upper tester

        Returns:
            Tester          Tester object created from the dictionary
        '''
        try:
            g = input_dict
            t = input_dict['tester']

            env = tricot.utils.merge_environment(t.get('env'), environment, path)
            error_mode = t.get('error_mode') or e_mode
            conds = Condition.from_dict(path, t.get('conditionals', {})).union(conditionals)
            run_conds = t.get('conditions', {})
            output_c = tricot.utils.merge(output_conf, t.get('output', {}), 'output', path)
            id_pattern = tricot.utils.verify_id_pattern(t.get('id_pattern'), path)

            Condition.check_format(path, run_conds, conds)

            testers = g.get('testers')
            definitions = g.get('tests')
            includes = g.get('include')

            groups = tricot.utils.merge_groups(test_groups, list(map(lambda x: str(x), t.get('groups', []))))

            variables = tricot.utils.merge(initial_vars, g.get('variables', {}), 'variables', path)
            variables['cwd'] = path.parent
            variables = tricot.utils.apply_variables(variables, copy.deepcopy(variables))

            plugins = Plugin.from_list(path, g.get('plugins'), variables)
            containers = TricotContainer.from_list(g.get('containers', list()), path, variables)

            tester_list = list()
            if testers and type(testers) is list:

                for f in testers:

                    if not Path(f).is_absolute():
                        f = path.parent.joinpath(f)

                    testers_to_add = []
                    for ff in sorted(glob.glob(str(f))):

                        if Path(ff).is_file() and (ff.endswith('.yml') or ff.endswith('.yaml')):
                            tester = Tester.from_file(ff, variables, None, error_mode, env, conds, output_c, groups)
                            testers_to_add.append(tester)

                    testers_to_add.sort(key=lambda x: x.id)
                    tester_list += testers_to_add

            tests = []

            if definitions and type(definitions) is list:
                tests += Test.from_list(path, definitions, variables, error_mode, env, conds, output_c, groups, id_pattern)

            if includes and type(includes) is list:

                for include in includes:
                    tests += Test.from_file(path.parent / Path(include), variables, error_mode, env, conds, output_c, groups,
                                            id_pattern, len(tests))

            if not tests and not tester_list:
                raise TesterKeyError('tests', path, optional='testers')

            tests = tests if tests else None

            new_tester = Tester(path, t['title'], variables, tests, tester_list, containers, plugins,
                                run_conds, conds, error_mode, t.get('id'), groups, t.get('requires'))
            new_tester.set_logfile(t.get('logfile'))

            return new_tester

        except KeyError as e:
            raise TesterKeyError(str(e), path)

    def set_logfile(self, logfile: str) -> None:
        '''
        Sets the logfile attribute on the Tester object.

        Parameters:
            logfile         File system path to the logfile

        Returns:
            None
        '''
        if logfile is None:
            return

        logfile = tricot.utils.apply_variables(logfile, self.variables)
        self.logfile = open(logfile, 'w')

    def from_file(filename: str, initial_vars: dict[str, Any] = dict(), runtime_vars: dict[str, Any] = None,
                  error_mode: str = None, env: dict = {}, conditionals: set[Condition] = set(),
                  output_conf: dict = {}, test_groups: list[list[str]] = []) -> Tester:
        '''
        Creates a new Tester object from a .yml file. The .yml file obviously needs to be in the
        expected format and requires all keys that are needed to construct a Tester object.

        Parameters:
            filename        File system path to the corresponding .yml file
            initial_vars    Additional variables. This is used internally when nesting Testers
            runtime_vars    Runtime variables (only specifiable when using tricot as library)
            error_mode      Current error mode setting
            env             Current environment variables
            conditionals    Conditions inherited from the previous tester
            output_conf     Output configuration inherited from upper tester
            test_groups     List of test groups inherited from the upper tester

        Returns:
            Tester          Tester object created from the file
        '''
        with open(filename, 'r') as f:

            try:
                config_dict = yaml.safe_load(f.read())

            except (yaml.parser.ParserError, yaml.scanner.ScannerError) as e:
                raise ExceptionWrapper(e, Path(filename))

        if '$env' not in initial_vars:
            tricot.utils.add_environment(initial_vars)

        if runtime_vars is not None and '$runtime' not in initial_vars:
            initial_vars['$runtime'] = runtime_vars

        return Tester.from_dict(config_dict, initial_vars, Path(filename), error_mode, env, conditionals,
                                output_conf, test_groups)

    def contains_id(self, t_ids: set[str]) -> bool:
        '''
        Checks whether the specified Test / Tester IDs are contained within this tester.

        Parameters:
            t_ids           Test / Tester IDs to look for

        Returns:
            bool            True if one of the ids is contained within the tester
        '''
        if not t_ids or self.runall:
            return True

        if self.id and {self.id}.issubset(t_ids):
            self.runall = True
            return True

        if self.tests:

            for test in self.tests:
                if test.id and {test.id}.issubset(t_ids):
                    return True

        for tester in self.testers:
            if tester.contains_id(t_ids):
                return True

        return False

    def contains_group(self, t_groups: list[list[str]]) -> bool:
        '''
        Checks whether the specified Group set exists within the tester.

        Parameters:
            t_groups        List of group lists to check in

        Returns:
            bool            True if set of groups is contained within the tester
        '''
        if not t_groups or self.runall:
            return True

        if tricot.utils.groups_contain(t_groups, self.groups):
            self.runall = True
            return True

        if self.tests:

            for test in self.tests:
                if tricot.utils.groups_contain(t_groups, test.groups):
                    return True

        for tester in self.testers:
            if tester.contains_group(t_groups):
                return True

        return False

    def skip_test(self, exclude: set[str], exclude_groups: list[list[str]]) -> bool:
        '''
        Checks whether the current test is contained within the exclude lists.

        Parameters:
            exclude             Set of Test / Tester IDs to exclude
            exclude_groups      List of group lists to exclude

        Returns:
            bool
        '''
        if exclude and self.id in exclude:
            return True

        elif exclude_groups and tricot.utils.groups_contain(exclude_groups, self.groups):
            return True

        return False

    def check_requirements(self):
        '''
        Checks whether all requirements for the current tester are satisfied. Since a
        successful run also requires successful sub-testers, sub-tester requirements are
        also checked.

        Parameters:
            None

        Returns:
            None
        '''
        if not self.requires or type(self.requires) is not dict:
            return

        requires = tricot.utils.apply_variables(self.requires, self.variables)

        for file in requires.get('files', []):

            if type(file) is str and not Path(file).exists():
                raise ExceptionWrapper(TricotRequiredFile(file), self.path)

            elif type(file) is dict:

                if not tricot.utils.file_integrity(file):
                    raise ExceptionWrapper(TricotRequiredFile(file.get('filename') + " (wrong hash value)"), self.path)

        for command in requires.get('commands', []):

            if which(command) is None:
                raise ExceptionWrapper(TricotRequiredCommand(command), self.path)

        version = requires.get('tricot')
        if not tricot.utils.match_version(version):

            message = ''
            for item in ['lt', 'le', 'eq', 'gt', 'ge']:

                ver = version.get(item)

                if ver is not None:
                    message += f'{item}: {ver} '

            raise ExceptionWrapper(TricotVersionMismatch(message), self.path)

        for tester in self.testers:
            tester.check_requirements()

    def filter(self, ids: set[str], groups: list[list[str]], exclude: set[str],
               exclude_groups: list[list[str]]) -> bool:
        '''
        Walks down the tester tree and removes all tests / testers that should be excluded
        by the current test configuration. If filtering is desired, this should be run before
        the tester.run method. The return value indicates whether the current tester has
        tests / testers left after the filtering has applied.

        Parameters:
            ids                 Set of Test / Tester IDs to run
            groups              List of group lists to run
            exclude             Set of Test / Tester IDs to exclude
            exclude_groups      List of group lists to exclude

        Returns:
            bool                Whether or not the tester has tests / testers left
        '''
        if self.skip_test(exclude, exclude_groups):
            return False

        elif not self.contains_id(ids) or not self.contains_group(groups):
            return False

        elif tricot.skip_until:

            if tricot.skip_until == self.id:
                tricot.skip_until = None

        if self.runall:
            self.filter_tests(None, None, exclude, exclude_groups)
            self.filter_testers(None, None, exclude, exclude_groups)

        else:
            self.filter_tests(ids, groups, exclude, exclude_groups)
            self.filter_testers(ids, groups, exclude, exclude_groups)

        if not self.tests and not self.testers:
            return False

        return True

    def filter_testers(self, ids: set[str], groups: list[list[str]], exclude: set[str],
                       exclude_groups: list[list[str]]) -> None:
        '''
        Iterates the tester array and removes all testers that should not be included by
        the current filter configuration.

        Parameters:
            ids                 Set of Test / Tester IDs to run
            groups              List of group lists to run
            exclude             Set of Test / Tester IDs to exclude
            exclude_groups      List of group lists to exclude

        Returns:
            None
        '''
        if not self.testers:
            return

        for tester in list(self.testers):
            if not tester.filter(ids, groups, exclude, exclude_groups):
                self.testers.remove(tester)

    def filter_tests(self, ids: set[str], groups: list[list[str]], exclude: set[str],
                     exclude_groups: list[list[str]]) -> None:
        '''
        Iterates the test array and removes all tests that should not be included by the
        current filter configuration.

        Parameters:
            ids                 Set of Test / Tester IDs to run
            groups              List of group lists to run
            exclude             Set of Test / Tester IDs to exclude
            exclude_groups      List of group lists to exclude

        Returns:
            None
        '''
        if not self.tests:
            return

        for test in list(self.tests):

            if tricot.skip_until:

                if tricot.skip_until == test.id:
                    tricot.skip_until = None

                else:
                    self.tests.remove(test)
                    continue

            if test.skip_test(exclude, exclude_groups):
                self.tests.remove(test)

            elif ids and not {test.id}.issubset(ids):
                self.tests.remove(test)

            elif groups and not tricot.utils.groups_contain(groups, test.groups):
                self.tests.remove(test)

    def run(self, hotplug_variables: dict[str, Any] = dict()) -> None:
        '''
        Runs the tester. Iterates over the tests and tester array and runs all
        tests / testers inside of it.

        Parameters:
            hotplug_variables   Hotplug variables to use during the test

        Returns:
            None
        '''
        if not self.tests and not self.testers:
            Logger.print_mixed_yellow('Skipping empty test:', self.title)
            return

        if not Condition.check_conditions(self.conditions, self.conditionals):
            Logger.print_mixed_yellow('Skipping test:', self.title)
            return

        tricot.Logger.print('')
        Logger.add_logfile(self.logfile)
        Logger.print_mixed_yellow('Starting test:', self.title, end=' ')

        if self.id and self.id != self.title:
            Logger.print_plain_blue(f'[{self.id}]')

        else:
            print()

        hotplug = hotplug_variables.copy()
        Logger.increase_indent()

        for plugin in self.plugins:
            plugin._run(hotplug)

        for container in self.containers:
            container.start_container()
            hotplug = {**hotplug, **container.get_container_variables()}

        self.run_tests(hotplug)
        self.run_childs(hotplug)

        for container in self.containers:
            container.stop_container()

        for plugin in self.plugins:
            plugin._stop()

        Logger.remove_logfile(self.logfile)
        Logger.decrease_indent()

    def run_tests(self, hotplug_variables: dict[str, Any]) -> None:
        '''
        Wrapper function that executes the tests specified in a tester.

        Parameters:
            hotplug_variables   Hotplug variables to use during the test

        Returns:
            None
        '''
        if not self.tests:
            return

        Logger.print('')

        for ctr in range(len(self.tests)):
            self.tests[ctr].run(f'{ctr+1}.', hotplug_variables)

    def run_childs(self, hotplug_variables: dict[str, Any]) -> None:
        '''
        Runs the child testers of the current tester.

        Parameters:
            hotplug_variables   Hotplug variables to use during the test

        Returns:
            None
        '''
        for tester in self.testers:

            try:
                tester.run(hotplug_variables)

            except tricot.PluginException as e:

                tricot.Logger.print("")

                if self.error_mode == 'break':
                    raise e

                else:
                    tricot.Logger.print_mixed_yellow('Caught', 'PluginException', 'from', end='', e=True)
                    tricot.Logger.print_mixed_blue_plain('', e.name, 'plugin in', end=' ')
                    tricot.Logger.print_yellow_plain(e.path.absolute())
                    tricot.Logger.print_mixed_blue('Original exception:', f'{type(e.original).__name__} - {e.original}')
                    tricot.Logger.print_blue('Tester is skipped.', e=True)
                    tricot.Logger.decrease_indent()

                    tricot.constants.LAST_ERROR = tricot.constants.PLUGIN_EXCEPTION

    def increase() -> None:
        '''
        Helper function that increases the global tester count and prints an offset
        if the current tester is not the first one.
        '''
        if Tester.tester_count != 0:
            Logger.print('')

        Tester.tester_count += 1



from __future__ import annotations

import re
import sys
import os.path
from typing import Any
from pathlib import Path

import tricot
import tricot.utils
from tricot.command import Command


this = sys.modules[__name__]
this.extractors = {}


def register_extractor(extractor_name: str, extractor_class: type) -> None:
    '''
    Registers an extractor class under the specified name. This function needs to
    be called to make extractors available within of .yml files. Some default
    extractor classes are registerd at the end of this file. Others can be manually
    added.

    Parameters:
        extractor_name      Name of the extractor to use in .yml files
        extractor_class     Reference to the corresponding extractor class

    Returns:
        None
    '''
    this.extractors[extractor_name] = extractor_class


def get_extractor(path: Path, extractor_name: str, param: Any, variables: dict[str:str]) -> Extractor:
    '''
    Searches for the specified extractor name within the registred extractors and
    uses the corresponding class to create an instance for the extractor. This instance
    is returned by the function. If the specified extractor name is not found, the
    function raises an ExtractorError.

    Parameters:
        path                Path to the .yml file that contains the extractor
        extractor_name      Name of the requested extractor
        param               Params to initialize the extractor with
        variables           Variables to initialize the extractor with
    '''
    ext_class = this.extractors.get(extractor_name)

    if ext_class is None:
        raise ExtractorError(path, f"Unable to find specified extractor '{extractor_name}'.")

    return ext_class(path, extractor_name, param, variables)


def get_extractor_list() -> list[str]:
    '''
    Returns a list of currently registered extractor names.

    Parameters:
        None

    Returns:
        extractor_list      List of registred extractors
    '''
    keys = this.extractors.keys()
    return list(keys)


class ExtractException(Exception):
    '''
    Extractors can throw this exception on failing to extract a value. Extractors can be configured
    to silently continue in this case or to throw an error (the corresponding test where the extractor
    was defined in just fails then).
    '''
    def __init__(self, message: str, ext: Extractor) -> None:
        '''
        Just add two additional attributes that are the path to the config file where the extractor is
        defined in and the extractor that caused the exception.
        '''
        self.extractor = ext
        super().__init__(message)


class ExtractorError(Exception):
    '''
    ExtractorErrors are raised when an Extractor was specified with incorrect
    parameters. The corresponding Exception just contains a message describing
    the problem.
    '''
    def __init__(self, path: Path, message: str) -> None:
        '''
        Just add one additional attribute that is the path to the config file where the extractor is
        defined in.
        '''
        self.path = str(path.resolve())
        super().__init__(message)


class Extractor:
    '''
    Extractors can be used within tests to extract parts of the commands output into variables.
    An example could look like this:

        - title: Example Test
          description: |-
              Just an example test

          command:
            - cat
            - /etc/passwd

          extractors:
              - regex:
                  pattern: ':(.+)$'
                  variable: shells
                  default: 'failed'

              - regex:
                  pattern: '^([^:]+):'
                  variable: users
                  default:
                    users: 'failed'
                    users-0: 'failed'
                    users-1: 'failed'

    The assignment of matched values to variables can be handeled differently for each extractor.
    Read the corresponding extractors documentation to get more information onto that.
    '''
    param_type = None
    inner_types = None

    def __init__(self, path: Path, name: str, param: Any, variables: dict[str, Any]) -> None:
        '''
        Initializes the Extractor.

        Parameters:
            path        Path to the .yml file that contains the extractor
            name        Name of the extractor (as specified in the .yml file)
            param       Parameters of the extractor (can be str, dict, int, bool, ...)
            variables   Variables contained in this dict are replaced in all str parameters
        '''
        self.path = path
        self.name = name
        self.param = param
        self.variables = variables

        self.command = None
        self.failure_string = None
        self.failure_color = None

        self.on_miss = self.param.get('on_miss', 'continue')
        self.check_param_type()

    def check_param_type(self) -> None:
        '''
        Checks whether the specified parameter type matches the expected one. If this is not
        the case, an ExtactorError is raised.

        Parameters:
            None

        Returns:
            None
        '''
        self.param = tricot.utils.apply_variables(self.param, self.variables)

        if self.param_type is None:
            return

        if type(self.param) is not self.param_type:
            message = f"Extractor '{self.name}' requires a parameter type of {str(self.param_type)}"
            raise ExtractorError(self.path, message)

        self.check_inner_types()
        self.check_streams()

    def check_inner_types(self) -> None:
        '''
        Extractors that require a dictionary or list types can specify the static class variable 'inner_types'
        to specify which inner types are expected within the collection types. When a list is used,
        'inner_types' is expected to contain a list with allowed types. In case of a dict, 'inner_types'
        is expected to contain a dict, containing the possible parameters as keys and a dict of
        {'required': bool, 'type': type} as value.

        The check is currently only implemented on the first layer (non recursive).

        Parameters:
            None

        Returns:
            None
        '''
        if type(self.param) is dict and type(self.inner_types) is dict:

            for key, value in self.inner_types.items():

                param = self.param.get(key)
                if value['required'] and param is None and not self.check_alternative(key):

                    message = f"Extractor '{self.name}' requires key with name '{key}' and type {value['type']}."
                    raise ExtractorError(self.path, message)

                if param is not None and type(param) is not value['type']:

                    message = f"Extractor '{self.name}' expects type {value['type']} for the '{key}' key."
                    raise ExtractorError(self.path, message)

        elif type(self.param) is list and type(self.inner_types) is list:

            for param in self.param:
                if type(param) not in self.inner_types:
                    message = f"Extractor '{self.name}' requires a parameter type of list[{str(self.inner_types)}."
                    raise ExtractorError(self.path, message)

        self.check_expected()

    def check_alternative(self, key: str) -> bool:
        '''
        Checks whether the specified key has alternative values defined. If so, the function
        checks whether one of these is contained within the parameter dictionary.

        Parameters:
            key         Key to look for

        Returns:
            bool        True if one of the alternatives of key is contained
        '''
        alternatives = self.inner_types[key].get('alternatives', [])

        for item in alternatives:
            if item in self.param:
                return True

        return False

    def check_expected(self) -> None:
        '''
        For Extractors that accept dictionaries, it is possible that typos mess up the test configuration.
        This function validates (for extractors using the inner_types attribute), that there are no keys
        that are not contained within the inner_types dict.

        Parameters:
            None

        Returns:
            None
        '''
        if type(self.param) is dict and type(self.inner_types) is dict:

            expected_keys = set(self.inner_types.keys())
            expected_keys.add('stream')
            expected_keys.add('output')
            expected_keys.add('on_miss')

            for key in self.param.keys():

                if key != 'description' and key not in expected_keys:
                    tricot.logging.Logger.print_yellow('Warning:', end=' ')
                    tricot.logging.Logger.print_mixed_blue_plain('Extractor', self.name, 'contains unexpected key', end=': ')
                    tricot.logging.Logger.print_yellow_plain(key, end=' ')
                    print(f'({self.path})')

    def from_list(path: Path, input_list: dict, variables: dict[str, Any] = {}) -> list[Extractor]:
        '''
        Constructs a new list of Extractors from a python list. The list content is usually
        the specified extractor configuration within a .yml file. Notice that it is tempting
        to use only a dictionary for extractors, as the list seems to be unecessary in most cases.
        However, when you want to use the same extractor twice for one test, you need a list.

        Parameters:
            path        Path to the .yml file that contains the extractor definition
            input_list  Python list as read in from a .yml file
            variables   Global variables that should be inherited by the extractor

        Returns:
            Extractors   Newly constructed Extractor objects
        '''
        if type(input_list) is dict:
            raise ExtractorError(path, "Extractors need to be specified as a list.")

        extractors = list()

        for item in input_list:
            for key, value in item.items():
                val = get_extractor(path, key, value, variables)
                extractors.append(val)

        return extractors

    def resolve_path(self, path: str) -> str:
        '''
        Resolves the specified path relative to the current extractor definition (file location of
        the extractors .yml file). If the input path is already absolute, it is just returned.

        Parameters:
            path        File system path to resolve

        Returns:
            resolved    Resolved file system path
        '''
        if os.path.isabs(path):
            return path

        return str(self.path.parent.joinpath(path).resolve())

    def check_streams(self) -> None:
        '''
        Used to validate stream names if specified. Is run during parsing of tests to prevent
        errors at runtime. Stream names can be specified for each extractor that accepts parameters
        as dictionary values.

        Parameters:
            None

        Returns:
            None
        '''
        if type(self.param) != dict:
            return

        stream = self.param.get('stream')

        if stream is not None:

            if type(stream) != str or stream not in ['stdout', 'stderr', 'both']:
                raise ExtractorError(self.path, "When specified, stream needs to be one of 'stdout', 'stderr' or 'both'.")

    def get_output(self) -> str:
        '''
        Helper function to obtain the actual output of a command execution. Each extractor accepts
        the special key 'stream' within of it's parameters. 'stream' is expected to be either 'both',
        'stdout' or 'stderr'. If it was not specified 'both' is taken as the default. The return value
        of this function is the command output that was captured for the correspoding stream.

        Parameters:
            None

        Returns:
            output      Output captured from stdout, stderr or both
        '''
        if type(self.param) != dict:
            return self.command.get_output()

        stream = self.param.get('stream')

        if stream == 'stdout':
            return self.command.stdout

        elif stream == 'stderr':
            return self.command.stderr

        elif stream == 'both' or stream is None:
            return self.command.get_output()

        raise ExtractorError(self.path, f"Encountered unexpected value for the 'stream' key: {stream}")

    def _extract(self, command: Command, hotplug_variables: dict[str, Any] = None) -> None:
        '''
        This method is called internally to perform the extraction process. It is basically
        a wrapper to the user defined 'extract' method that stores the command result in an
        additional class variable. All exceptions thrown by the actual extractor are encapsulated
        within an ExtractException.

        Parameters:
            command             Command object associated with the current tests
            hotplug_variables   Variables to apply to the extractor at runtime

        Returns:
            None
        '''
        self.command = command
        self.param = tricot.utils.apply_variables(self.param, hotplug_variables)

        try:
            self.extract(hotplug_variables)

        except ExtractException as e:
            raise e

        except Exception as e:
            raise ExtractException(str(e), self)

    def extract(self, hotplug: dict) -> None:
        '''
        Dummy extract method. This one needs to be overwritten by the actual extractor implementations.

        Parameters:
            hotplug             Hotplug variables. Required by the extractor to add new variables.

        Returns:
            None
        '''
        return None


class RegexExtractor(Extractor):
    '''
    The RegexExtractor extracts command output based on a regular expression. The following yaml
    shows an example:

        - title: Example Test
          description: |-
              Just an example test

          command:
            - cat
            - /etc/passwd

          extractors:
              - regex:
                  pattern: ':(.+)$'
                  variable: shells

              - regex:
                  pattern: '^([^:]+):'
                  variable: users
                  default:
                    users: 'failed'
                    users-0: 'failed'
                    users-1: 'failed'

    The extraction results of this extractor can be accessed like variables. It support
    multiple matches and also match groups. The literal variable name gets always assigned the
    first matching item. Additionally, you can access matching items by using 'variablename-#',
    where '#' is the desired match number. When using match groups, you can access matches from
    different groups by using 'variablename-#-#', where the first '#' is the number of the match
    group and the second '#' is the number of the match.
    '''
    param_type = dict
    inner_types = {
            'ascii': {'required': False, 'type': bool},
            'default': {'required': False, 'type': dict},
            'dotall': {'required': False, 'type': bool},
            'ignore_case': {'required': False, 'type': bool},
            'multiline': {'required': False, 'type': bool},
            'pattern': {'required': True, 'type': str},
            'variable': {'required': True, 'type': str},
    }

    def __init__(self, *args, **kwargs) -> None:
        '''
        Regex initializer. This is used to throw ExtractorErrors in case of invalid
        regex specifications already on creation time.
        '''
        super().__init__(*args, **kwargs)

        flags = 0

        if self.param.get('ascii'):
            flags = flags | re.ASCII
        if self.param.get('dotall'):
            flags = flags | re.DOTALL
        if self.param.get('ignore_case'):
            flags = flags | re.IGNORECASE
        if self.param.get('multiline'):
            flags = flags | re.MULTILINE

        try:
            self.pattern = self.param.get('pattern', '')
            self.regex = re.compile(self.pattern, flags)
            self.variable = self.param.get('variable')
            self.defaults = self.param.get('default', {})

        except Exception:
            raise ExtractorError(self.path, f"Specified regex '{self.pattern}' is invalid!")

    def extract(self, hotplug: dict) -> None:
        '''
        Extract parts of the command output into variables.
        '''
        cmd_output = self.get_output()

        matches = self.regex.finditer(cmd_output)

        for key, value in self.defaults.items():
            hotplug[key] = value

        ctr = 0
        for match in matches:

            if ctr == 0:
                hotplug[self.variable] = match.group(0)

            key = f'{self.variable}-{ctr}'
            hotplug[key] = match.group(0)

            key = f'{self.variable}-{ctr}-0'
            hotplug[key] = match.group(0)

            cts = 1

            for group in match.groups():

                key = f'{self.variable}-{ctr}-{cts}'
                hotplug[key] = match.group(cts)
                cts += 1

            ctr += 1

        if ctr == 0:
            raise ExtractException(f"RegexExtractor did not find pattern '{self.pattern}' within the output.", self)


register_extractor("regex", RegexExtractor)



from __future__ import annotations

import os
import re
import tricot
import hashlib
from typing import Any
from pathlib import Path


class TricotRuntimeVariableError(Exception):
    '''
    A TricotRuntimeVariableError is raised, when a runtime variable was specified within
    the .yml test configuration, but was not defined at runtime.
    '''


class TricotEnvVariableError(Exception):
    '''
    A TricotEnvVariableError is raised, when an environment variable was specified within
    the .yml test configuration, but was not found at runtime.
    '''


class TricotInvalidVersionException(Exception):
    '''
    Is raised when an invalid version string was specified.
    '''


def resolve_runtime_variables(variables: dict[str, Any], key: str, value: Any) -> Any:
    '''
    This function is responsible for resolving runtime variables.
    When a variable contains the value '$runtime' it checks the
    '$runtime' key within the variables dict for the corresponding
    dictionary key. If it is found, the corresponding value is
    returned. If not, a TricotRuntimeVariableError is raised.

    When the value of a variables is not '$runtime', it is directly
    returned.

    Parameters:
        variables       Variables dictionary created during the Tester initialization
        key             Key of the current variable in question
        value           Value of the current variable in question

    Returns:
        value       Value with possible runtime variables applied
    '''
    if value != '$runtime':
        return value

    try:
        return (variables['$runtime'])[key]

    except KeyError:
        raise TricotRuntimeVariableError(f"Unable to find runtime variable '{key}'.")


def resolve_env_variables(variables: dict[str, Any], key: str, value: Any) -> Any:
    '''
    This function is responsible for resolving environment variables.
    When a variable contains the value '$env' it checks the
    '$env' key within the variables dict for the corresponding
    dictionary key. If it is found, the corresponding value is
    returned. If not, a TricotEnvVariableError is raised.

    When the value of a variables is not '$env', it is directly
    returned.

    Parameters:
        variables       Variables dictionary created during the Tester initialization
        key             Key of the current variable in question
        value           Value of the current variable in question

    Returns:
        value       Value with possible runtime variables applied
    '''
    if value != '$env':
        return value

    try:
        return (variables['$env'])[key]

    except KeyError:
        raise TricotEnvVariableError(f"Unable to find environment variable '{key}'.")


def apply_variables(candidate: Any, var_dict: dict[str, Any]) -> Any:
    '''
    Applies the variables contained in the Plugin on each str value within the
    parameters. For dictionaires and list, the function uses recursion to iterate
    over all possible items.

    Parameters:
        candidate       Current param value. Required for recursion.
        var_dict        Variables to apply.

    Returns:
        None
    '''
    if candidate is None or var_dict is None:
        return candidate

    cur_type = type(candidate)

    if cur_type is dict:
        for key, value in candidate.items():
            candidate[key] = apply_variables(value, var_dict)

    elif cur_type is list:
        for ctr in range(len(candidate)):
            candidate[ctr] = apply_variables(candidate[ctr], var_dict)

    elif cur_type is str:

        for var, var_value in var_dict.items():

            var_value = resolve_runtime_variables(var_dict, var, var_value)
            var_value = resolve_env_variables(var_dict, var, var_value)
            variable_key = '${'+str(var)+'}'

            if candidate == variable_key:
                candidate = var_value

                if type(candidate) is not str:
                    break

            else:
                candidate = candidate.replace(variable_key, str(var_value))

    return candidate


def make_ordinal(n: int) -> str:
    '''
    Convert an integer into its ordinal representation. Used for pretty printing and copied
    from: https://stackoverflow.com/questions/9647202/ordinal-numbers-replacement

        make_ordinal(0)   => '0th'
        make_ordinal(3)   => '3rd'
        make_ordinal(122) => '122nd'
        make_ordinal(213) => '213th'
    '''
    n = int(n)
    suffix = ['th', 'st', 'nd', 'rd', 'th'][min(n % 10, 4)]
    if 11 <= (n % 100) <= 13:
        suffix = 'th'
    return str(n) + suffix


def check_keys(expected_keys: list[str], yaml_dict: dict) -> None:
    '''
    What happes quite some time is that validators are indented incorrectly and appear within the test
    section of the .yml definition. This is annoying, as your test seems to work, but was actually never
    validated. This function checks for unexpcted keys within .yml files and prints a warning if one is
    encountered.

    Parameters:
        expected_keys   List of keys that are expected to appear
        yaml_dict       Python dirctionary that represents the .yml file

    Returns:
        None
    '''
    for key in yaml_dict.keys():
        if key not in expected_keys:
            tricot.Logger.print_yellow('Warning:', end=' ')
            tricot.Logger.print_mixed_blue_plain('Test', yaml_dict['title'], 'contains unexpected key', end=': ')
            tricot.Logger.print_yellow_plain(key)


def merge_environment(new: dict, current: dict, path: Path) -> dict:
    '''
    Helper function to merge two environment dictionaries. Is mainly used to handle
    the case where the new environment is not a dictionary.

    Parameters:
        new             New environment variables to add
        current         Current set of environment variables
        path            Path of the currently parsed configuration file

    Returns:
        environment     Merged environment variables
    '''
    current = current or {}

    if new is None:
        return current

    if type(new) is not dict:
        raise tricot.TestKeyError(None, path, "Key 'env' needs to be a dictionary in the ")

    return {**new, **current}


def merge_default_environment(env: dict) -> dict:
    '''
    Merges the current user environment with the specifies environment dictionary.

    Parameters:
        new             dict to combine with the default environment

    Returns:
        environment     Merged environment variables
    '''
    return {**os.environ, **env}


def add_environment(variables: dict[str, Any]) -> None:
    '''
    Adds the current user environment to the available variables within the
    $env key.

    Parameters:
        variables       Variable dictionary

    Returns:
        None
    '''
    variables['$env'] = os.environ


def validate_color(string: str, allow_none: bool = False) -> None:
    '''
    Validates that the input string is a valid color name. If not, an exception is raised.

    Parameters:
        variables       Variable dictionary
        allow_none      Whether None values are allowed

    Returns:
        None
    '''
    if string is None and allow_none:
        return

    colors = ['grey', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white']
    if string not in colors:
        raise Exception(f'Invalid color was specified: {string}')


def validate_type(name: str, value: Any, expected_type: Any, path: Path = None) -> None:
    '''
    Validates that the input parameter 'value' has a type of 'expected_type'.
    If not, a corresponding exception is raised that contains the 'name' of the
    parameter and it's expected type

    Parameters:
        name            Name of the parameter
        value           Value of the parameter
        expected_type   Expected type of the parameter
        path            Optional path to a configuration file

    Returns:
        None
    '''
    if type(value) is not expected_type:
        raise tricot.TricotException(f"The '{name}' attribute requires type '{expected_type}'.", path)


def merge(dict1: dict, dict2: dict, name: str = None, path: Path = None) -> dict:
    '''
    Merge two dictionaries.

    Parameters:
        dict1       Dictionary one to merge
        dict2       Dictionary two to merge
        name        Optional string to print in exceptions
        path        Optional path to a configuration file

    Returns:
        dict        Merged dictionary
    '''
    if type(dict1) is not dict or type(dict2) is not dict:

        if name:
            raise tricot.TricotException(f"The '{name}' attribute requires type 'dict'.", path)

        else:
            raise tricot.TricotException('Invalid type specified during merge operation.', path)

    return {**dict1, **dict2}


def parse_groups(groups: list[str]) -> list[list[str]]:
    '''
    Parses group specifications. Groups should be specified as comma separated strings. Each
    comma separated part is interpreted as a group. All groups within a string are mandatory
    for a test / tester to match. Braces can be used for or-like statements.

    E.g.:

        java8,networking,filter      -> list(java8, networking, filter)
        java8,{networking,io},filter -> list(list(java8, networking, filter),
                                             list(java8, io, filter))

    Parameters:
        groups          List of group specifications

    Returns:
        list            List of group lists
    '''
    lists = list()
    regex = re.compile(r'\{([^}]+)\}')

    for group_spec in groups:

        or_like = regex.findall(group_spec)
        group_spec = regex.sub('$ORLIKE$', group_spec)

        split = list(filter(None, group_spec.split(',')))
        group_lists = [split]

        for match in or_like:

            new = []

            for group_list in group_lists:

                split = list(filter(None, match.split(',')))

                for item in split:

                    new_list = group_list.copy()

                    for ctr in range(len(new_list)):

                        if new_list[ctr] == '$ORLIKE$':
                            new_list[ctr] = item
                            break

                    new.append(new_list)

            group_lists = new

        lists += group_lists

    return lists


def groups_contain(groups_list: list[list[str]], groups: list[list[str]]) -> bool:
    '''
    Checks whether a specified list of groups contains a particular group of a
    list of specified groups. This separate function is required, as group
    comparison supports wildcards as * or **.

    Parameters:
        groups_list           List of group lists to search in
        groups                Group list to look for

    Returns:
        bool                  True if group is contained in groups
    '''
    for group in groups:

        for items in groups_list:

            ctr = 0
            match = True
            items = items.copy()

            try:

                while len(items) != 0:

                    item = items.pop(0)

                    if item == '*' or group[ctr] == item:

                        ctr += 1
                        continue

                    if item == '**':

                        item = items.pop(0)
                        while group[ctr] != item and ctr != len(group):
                            ctr += 1

                        if ctr == len(group):
                            match = False
                            break

                        ctr += 1
                        continue

                    match = False
                    break

                if match:
                    return True

            except IndexError:
                pass

    return False


def merge_groups(parent_groups: list[list[str]], new_groups: list[str]) -> list[list[str]]:
    '''
    This function is called by tests and testers to join groups that are defined within
    the test / tester definition with group lists that have been specified for upper testers.
    Each group in the test / tester specification is appened to the parent defined groups.

    Paramaters:
        parent_groups           Group lists inherited by the parent
        new_groups              Groups specified for the test / tester

    Returns:
        merged                  Merge result
    '''
    merged = list()

    for parent_group in (parent_groups or [[]]):

        if new_groups:

            for new_group in new_groups:
                copy = parent_group.copy()
                copy.append(new_group)
                merged.append(copy)

        else:
            merged.append(parent_group.copy())

    return merged


def verify_id_pattern(pattern: str, path: Path) -> str:
    '''
    Checks whether the specified ID pattern is valid. If invalid, a TricotException
    is raised.

    Parameters:
        pattern         ID pattern defined in a tester
        path            Path of the currently parsed configuration file

    Returns:
        str             If valid, the id pattern
    '''
    if pattern is not None:

        if type(pattern) is not str:
            raise tricot.TricotException("Tester attribute 'id_pattern' needs to be string.", path)

        regex = re.compile(r'\{(?::[0-9<>=^+bcdoxXneEfFgGn% -]+)?\}')
        matches = regex.findall(pattern)

        if len(matches) != 1:
            msg = "Tester attribute 'id_pattern' needs to contain exactly one '{}' sequence."
            raise tricot.TricotException(msg, path)

    return pattern


def match_version(version_dict: dict) -> bool:
    '''
    Checks if the currently running tricot installation matches the version specified
    in version_dict. version_dict is a dictionary that has to contain at least one of
    the following keys: eq, lt, gt.

    Paramaters:
        version_dict        Version dictionary containing at least one of eq, lt, gt

    Returns:
        bool                True if matching, False otherwise
    '''
    if version_dict is None:
        return True

    try:

        cur_ver = tricot.constants.VERSION

        for comperator in ['lt', 'le', 'eq', 'gt', 'ge']:

            version = version_dict.get(comperator)

            if version is None:
                continue

            result = compare_versions(cur_ver, version)

            if 'e' in comperator and result == 0:
                pass

            elif 'l' in comperator and result == -1:
                pass

            elif 'g' in comperator and result == 1:
                pass

            else:
                return False

    except TricotInvalidVersionException:
        return False

    return True


def split_version(version_string: str) -> list[int]:
    '''
    Splits a version string into it's three different components.

    Parameters:
        version_string          Incoming version string

    Returns:
        list[str]               List containing the different version components
    '''
    version_list = list()
    split = version_string.split('.')

    for item in split:

        if not item.isdigit():
            raise TricotInvalidVersionException(version_string)

        version_list.append(int(item))

    while len(version_list) < 3:
        version_list.append(0)

    return version_list


def compare_versions(one: str, other: str) -> int:
    '''
    Compares two version strings with each other. If equal,
    zero is reqturned. -1 if first is lower, +1 if other is lower.

    Parameters:
        one             Version string
        other           Version string

    Returns:
        int             0 -> equal, -1 -> one < other, +1 -> one > other
    '''
    split1 = split_version(one)
    split2 = split_version(other)

    for ctr in range(0, 3):

        if split1[ctr] < split2[ctr]:
            return -1

        elif split1[ctr] > split2[ctr]:
            return 1

    return 0


def file_integrity(file_dict: dict) -> bool:
    '''
    Checks whether the filename matches the specified checksums.
    The filename is expected under the 'filename' key of the specified
    dicitionary. Checksum keys can be md5, sha1, sha256, sha512

    Parameters:
        file_dict           Dictionary containing the filename and checksums

    Returns:
        bool                True if file exists and has the correct hash
    '''
    filename = file_dict.get('filename')

    if not filename or not Path(filename).exists():
        return False

    for hash_type in ['md5', 'sha1', 'sha256', 'sha512']:

        hash_value = file_dict.get(hash_type)

        if hash_value is not None:

            with open(filename, 'rb') as f:
                content = f.read()

            if hashlib.new(hash_type, content).hexdigest() != hash_value:
                return False

    return True
